@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2018 Fabian Groffen
+ * Copyright 2013-2020 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -22,6 +22,7 @@
 #include <errno.h>
 #include <poll.h>
 #include <pthread.h>
+#include <signal.h>
 #include <sys/uio.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -41,6 +42,7 @@
 #endif
 #ifdef HAVE_LZ4
 #include <lz4.h>
+#include <lz4frame.h>
 #endif
 #ifdef HAVE_SNAPPY
 #include <snappy-c.h>
@@ -49,6 +51,15 @@
 #include <openssl/ssl.h>
 #endif
 
+#if defined(HAVE_DISPATCH_DISPATCH_H)
+# include <dispatch/dispatch.h>
+#elif defined(HAVE_SEMAPHORE_H)
+# include <semaphore.h>
+#else
+# error "found no implementation for semaphores for your platform!"
+#endif
+
+
 enum conntype {
 	LISTENER,
 	CONNECTION
@@ -62,7 +73,7 @@ typedef struct _z_strm {
 		z_stream z;
 #endif
 #ifdef HAVE_LZ4
-		LZ4_streamDecode_t *lz;
+		LZ4F_decompressionContext_t lz;
 #endif
 #ifdef HAVE_SSL
 		SSL *ssl;
@@ -90,35 +101,39 @@ typedef struct _connection {
 	int buflen;
 	char needmore:1;
 	char noexpire:1;
+	char isaggr:1;
+	char isudp:1;
+	char datawaiting; /* full byte for atomic access */
 	char metric[METRIC_BUFSIZ];
 	destination dests[CONN_DESTS_SIZE];
 	size_t destlen;
 	struct timeval lastwork;
 	unsigned int maxsenddelay;
-	char hadwork:1;
-	char isaggr:1;
-	char isudp:1;
 } connection;
 
 struct _dispatcher {
 	pthread_t tid;
 	enum conntype type;
-	char id;
 	size_t metrics;
 	size_t blackholes;
+	size_t discards;
 	size_t ticks;
 	size_t sleeps;
 	size_t prevmetrics;
 	size_t prevblackholes;
+	size_t prevdiscards;
 	size_t prevticks;
 	size_t prevsleeps;
-	char keep_running;  /* full byte for atomic access */
+	char id;
+	char keep_running;  /* these all use a full byte for atomic access */
+	char route_refresh_pending;
+	char hold;
+	char tags_supported;
 	router *rtr;
 	router *pending_rtr;
-	char route_refresh_pending;  /* full byte for atomic access */
-	char hold;  /* full byte for atomic access */
 	char *allowed_chars;
-	char tags_supported;
+	int maxinplen;
+	int maxmetriclen;
 };
 
 static listener **listeners = NULL;
@@ -193,6 +208,7 @@ gzipread(z_strm *strm, void *buf, size_t sze)
 	int ret;
 	int iret;
 	int err = 0;
+	int inflatemode = Z_SYNC_FLUSH;
 
 	/* read any available data, if it fits */
 	ret = strm->nextstrm->strmread(strm->nextstrm,
@@ -202,11 +218,13 @@ gzipread(z_strm *strm, void *buf, size_t sze)
 		strm->ipos += ret;
 	} else if (ret < 0) {
 		err = errno;
+		if (err != EAGAIN)
+			inflatemode = Z_FINISH;
 	} else {
 		/* ret == 0: EOF, which means we didn't read anything here, so
-		 * calling inflate again will be pointless, because there is
-		 * nothing new. */
-		return 0;
+		 * calling inflate should be to flush whatever is in the zlib
+		 * buffers, much like a read error. */
+		inflatemode = Z_FINISH;
 	}
 
 	zstrm->next_in = (Bytef *)ibuf;
@@ -214,7 +232,7 @@ gzipread(z_strm *strm, void *buf, size_t sze)
 	zstrm->next_out = (Bytef *)buf;
 	zstrm->avail_out = (uInt)sze;
 
-	iret = inflate(zstrm, Z_SYNC_FLUSH);
+	iret = inflate(zstrm, inflatemode);
 
 	/* if inflate consumed something, update our ibuf */
 	if ((Bytef *)ibuf != zstrm->next_in) {
@@ -279,37 +297,78 @@ static inline ssize_t
 lzread(z_strm *strm, void *buf, size_t sze)
 {
 	char *ibuf = strm->ibuf;
+	size_t srcsize;
+	size_t destsize;
 	int ret;
 
 	/* read any available data, if it fits */
 	ret = strm->nextstrm->strmread(strm->nextstrm,
 			ibuf + strm->ipos, METRIC_BUFSIZ - strm->ipos);
+
+	/* if EOF(0) or no-data(-1) then only get out now if the input
+	 * buffer is empty because start of next frame may be waiting for us */
+	
 	if (ret > 0) {
 		strm->ipos += ret;
 	} else if (ret < 0) {
-		return -1;
+		if (strm->ipos == 0)
+			return -1;
 	} else {
 		/* ret == 0, a.k.a. EOF */
-		return 0;
+		if (strm->ipos == 0)
+			return 0;
 	}
 
-	ret = LZ4_decompress_safe_continue(strm->hdl.lz,
-			ibuf, buf, strm->ipos, sze);
+	/* attempt to decompress something from the (partial) frame that's
+	 * arrived so far.  srcsize is updated to the number of bytes
+	 * consumed. likewise for destsize and bytes written */
+
+	srcsize = strm->ipos;
+	destsize = sze;
+	
+	ret = LZ4F_decompress(strm->hdl.lz, buf, &destsize, ibuf, &srcsize, NULL);
+
+	/* check for error before doing anything else */
+
+	if (LZ4F_isError(ret)) {
+
+		/* liblz4 doesn't allow access to the error constants so have to
+		 * return a generic code */
+
+		logerr("Error %d reading LZ4 compressed data\n", (int)ret);
+		errno = EBADMSG;
+		strm->ipos = 0;
+
+		return -1;
+	}
 
 	/* if we decompressed something, update our ibuf */
-	if (ret > 0) {
-		strm->ipos = strm->ipos - ret;
-		memmove(ibuf, ibuf + ret, strm->ipos);
+
+	if (srcsize > 0) {
+		memmove(ibuf, ibuf + srcsize, strm->ipos - srcsize);
+		strm->ipos -= srcsize;
 	}
 
-	return (ssize_t)ret;
+	if (destsize == 0) {
+		tracef("No LZ4 data was produced\n");
+		errno = EAGAIN;
+		return -1;
+	}
+
+#ifdef ENABLE_TRACE
+	/* debug logging */
+	if (ret == 0)
+		tracef("LZ4 frame fully decoded\n");
+#endif
+
+	return (ssize_t)destsize;
 }
 
 static inline int
 lzclose(z_strm *strm)
 {
 	int ret = strm->nextstrm->strmclose(strm->nextstrm);
-	LZ4_freeStreamDecode(strm->hdl.lz);
+	LZ4F_freeDecompressionContext(strm->hdl.lz);
 	free(strm);
 	return ret;
 }
@@ -443,7 +502,7 @@ dispatch_addlistener(listener *lsnr)
 	}
 	if (c == MAX_LISTENERS) {
 		logerr("cannot add new listener: "
-				"no more free listener slots (max = %zu)\n",
+				"no more free listener slots (max = %d)\n",
 				MAX_LISTENERS);
 		pthread_rwlock_unlock(&listenerslock);
 		return 1;
@@ -493,8 +552,10 @@ dispatch_removelistener(listener *lsnr)
 		*socks = -1;
 	}
 	pthread_rwlock_unlock(&connectionslock);
-	if (lsnr->saddrs)
+	if (lsnr->saddrs) {
 		freeaddrinfo(lsnr->saddrs);
+		lsnr->saddrs = NULL;
+	}
 }
 
 /**
@@ -564,6 +625,14 @@ dispatch_addconnection(int sock, listener *lsnr)
 
 			pthread_rwlock_unlock(&connectionslock);
 			return -1;
+		} else if (newlst != connections) {
+		    /* reset srcaddr after realloc due to issue 346 */
+		    for (c = 0; c < connectionslen; c++) {
+			if (newlst[c].isudp) {
+			    newlst[c].strm->hdl.udp.srcaddr = newlst[c].srcaddr;
+			    newlst[c].strm->hdl.udp.srcaddrlen = sizeof(newlst[c].srcaddr);
+			}
+		    }
 		}
 
 		for (c = connectionslen; c < connectionslen + CONNGROWSZ; c++) {
@@ -670,7 +739,11 @@ dispatch_addconnection(int sock, listener *lsnr)
 			__sync_bool_compare_and_swap(&(connections[c].takenby), -2, -1);
 			return -1;
 		}
-		lzstrm->hdl.lz = LZ4_createStreamDecode();
+		if (LZ4F_isError(LZ4F_createDecompressionContext(&lzstrm->hdl.lz, LZ4F_VERSION))) {
+			logerr("Failed to create LZ4 decompression context\n");
+			return -1;
+		}
+		
 		lzstrm->strmread = &lzread;
 		lzstrm->strmclose = &lzclose;
 		lzstrm->nextstrm = connections[c].strm;
@@ -699,7 +772,7 @@ dispatch_addconnection(int sock, listener *lsnr)
 	connections[c].isudp = 0;
 	connections[c].destlen = 0;
 	gettimeofday(&connections[c].lastwork, NULL);
-	connections[c].hadwork = 1;  /* force first iteration before stalling */
+	connections[c].datawaiting = 0;
 	/* after this dispatchers will pick this connection up */
 	__sync_bool_compare_and_swap(&(connections[c].takenby), -2, 0);
 	__sync_add_and_fetch(&acceptedconnections, 1);
@@ -745,7 +818,8 @@ dispatch_process_dests(connection *conn, dispatcher *self, struct timeval now)
 					self->id, conn->sock, conn->dests[i].metric,
 					server_ip(conn->dests[i].dest),
 					server_port(conn->dests[i].dest));
-			if (server_send(conn->dests[i].dest, conn->dests[i].metric, force) == 0)
+			if (server_send(conn->dests[i].dest,
+						conn->dests[i].metric, force) == 0)
 				break;
 		}
 		if (i < conn->destlen) {
@@ -757,7 +831,6 @@ dispatch_process_dests(connection *conn, dispatcher *self, struct timeval now)
 			/* finally "complete" this metric */
 			conn->destlen = 0;
 			conn->lastwork = now;
-			conn->hadwork = 1;
 		}
 	}
 
@@ -789,6 +862,7 @@ static int
 dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 {
 	char *p, *q, *firstspace, *lastnl;
+	char search_tags;
 	int len;
 
 	/* first try to resume any work being blocked */
@@ -797,13 +871,12 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 		return 0;
 	}
 
-	/* don't poll (read) when the last time we ran nothing happened,
-	 * this is to avoid excessive CPU usage, issue #126 */
-	if (!conn->hadwork && timediff(conn->lastwork, start) < 100 * 1000) {
+	/* only read from socket when there's actually something to read,
+	 * this should avoid excessive CPU usage, issue #126 */
+	if (__sync_bool_compare_and_swap(&(conn->datawaiting), 0, 0)) {
 		__sync_bool_compare_and_swap(&(conn->takenby), self->id, 0);
 		return 0;
 	}
-	conn->hadwork = 0;
 
 	len = -2;
 	/* try to read more data, if that succeeds, or we still have data
@@ -836,13 +909,19 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 		q = conn->metric;
 		firstspace = NULL;
 		lastnl = NULL;
+		search_tags = self->tags_supported;
 		for (p = conn->buf; p - conn->buf < conn->buflen; p++) {
 			if (*p == '\n' || *p == '\r') {
 				/* end of metric */
 				lastnl = p;
 
-				/* just a newline on it's own? some random garbage? skip */
-				if (q == conn->metric || firstspace == NULL) {
+				/* just a newline on it's own? some random garbage?
+				 * do we exceed the set limits? drop */
+				if (q == conn->metric || firstspace == NULL ||
+						q - conn->metric > self->maxinplen - 1 ||
+						firstspace - conn->metric > self->maxmetriclen)
+				{
+					__sync_add_and_fetch(&(self->discards), 1);
 					q = conn->metric;
 					firstspace = NULL;
 					continue;
@@ -870,8 +949,8 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 				/* restart building new one from the start */
 				q = conn->metric;
 				firstspace = NULL;
+				search_tags = self->tags_supported;
 
-				conn->hadwork = 1;
 				gettimeofday(&conn->lastwork, NULL);
 				conn->maxsenddelay = 0;
 				/* send the metric to where it is supposed to go */
@@ -899,8 +978,9 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 					if (*(q - 1) != *p && (q - 1) != firstspace)
 						*q++ = *p;
 				}
-			} else if (self->tags_supported && *p == ';') {
+			} else if (search_tags && *p == ';') {
 				/* copy up to next space */
+				search_tags = 0;
 				firstspace = q;
 				*q++ = *p;
 			} else if (
@@ -977,6 +1057,7 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 
 			/* flag this connection as no longer in use, unless there is
 			 * pending metrics to send */
+			conn->sock = -1;  /* ensure a poll won't match */
 			__sync_bool_compare_and_swap(&(conn->takenby), self->id, -1);
 
 			return 0;
@@ -993,6 +1074,23 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
  * pthread compatible routine that handles connections and processes
  * whatever comes in on those.
  */
+#ifdef HAVE_DISPATCH_DISPATCH_H
+static dispatch_semaphore_t *datawaiting = NULL;
+static dispatch_semaphore_t _datawaiting;
+#define sem_init(X,Y,Z) *(X) = dispatch_semaphore_create(Z)
+#define sem_post(X)            dispatch_semaphore_signal(*(X))
+#define sema_wait(X,T)         dispatch_semaphore_wait(*(X), \
+                                 dispatch_time(DISPATCH_TIME_NOW, T))
+#else
+static sem_t *datawaiting = NULL;
+static sem_t _datawaiting;
+static int sema_wait(sem_t *sem, int timeout) {
+	struct timespec wait;
+	wait.tv_sec = 0;
+	wait.tv_nsec = timeout;
+	return sem_timedwait(sem, &wait);
+}
+#endif
 static void *
 dispatch_runner(void *arg)
 {
@@ -1001,12 +1099,44 @@ dispatch_runner(void *arg)
 	int c;
 
 	if (self->type == LISTENER) {
-		struct pollfd ufds[MAX_LISTENERS];
+		struct pollfd *ufds = NULL;
+		int ufdslen = 0;
 		int fds;
+		int cfds;
+		int f;
 		int *sock;
+
 		while (__sync_bool_compare_and_swap(&(self->keep_running), 1, 1)) {
-			pthread_rwlock_rdlock(&listenerslock);
+			pthread_rwlock_rdlock(&connectionslock);
+			if (ufdslen < MAX_LISTENERS + connectionslen) {
+				ufdslen = MAX_LISTENERS + connectionslen;
+				ufds = realloc(ufds, sizeof(ufds[0]) * ufdslen);
+				if (ufds == NULL) {
+					logerr("dispatch: failed to allocate socket poll "
+							"space, cannot continue!\n");
+					pthread_rwlock_unlock(&connectionslock);
+					kill(getpid(), SIGTERM);
+					return NULL;
+				}
+			}
 			fds = 0;
+			for (c = 0; c < connectionslen; c++) {
+				conn = &(connections[c]);
+				if (!__sync_bool_compare_and_swap(&(conn->takenby), 0, 0))
+					continue;
+				/* connections are only read from if we flagged that
+				 * there is data waiting, so sockets cannot disappear
+				 * for the read code doesn't trigger unless we polled it */
+				if (__sync_bool_compare_and_swap(&(conn->datawaiting), 0, 0))
+				{
+					ufds[fds].fd = conn->sock;
+					ufds[fds].events = POLLIN | POLLERR;
+					fds++;
+				}
+			}
+			pthread_rwlock_unlock(&connectionslock);
+			cfds = fds;
+			pthread_rwlock_rdlock(&listenerslock);
 			for (c = 0; c < MAX_LISTENERS; c++) {
 				if (listeners[c] == NULL)
 					continue;
@@ -1017,9 +1147,33 @@ dispatch_runner(void *arg)
 				}
 			}
 			pthread_rwlock_unlock(&listenerslock);
-			if (poll(ufds, fds, 1000) > 0) {
-				int f;
-				for (f = fds - 1; f >= 0; f--) {
+			if (poll(ufds, fds, 1000) <= 0)
+				continue;
+
+			for (f = 0; f < fds; f++) {
+				if (f < cfds) {
+					/* connection has data available */
+					if (ufds[f].revents & (POLLIN | POLLERR)) {
+						pthread_rwlock_rdlock(&connectionslock);
+						for (c = 0; c < connectionslen; c++) {
+							conn = &(connections[c]);
+							/* connection may be serviced at this point,
+							 * that's fine */
+							if ((char)__sync_add_and_fetch(&(conn->takenby), 0)
+									< 0)
+								continue;
+							if (conn->sock == ufds[f].fd) {
+								__sync_bool_compare_and_swap(
+										&(conn->datawaiting), 0, 1);
+								sem_post(datawaiting);
+								tracef("data waiting on connection %d, "
+										"src %s\n", conn->sock, conn->srcaddr);
+							}
+						}
+						pthread_rwlock_unlock(&connectionslock);
+					}
+				} else {
+					/* listener has new connection */
 					if (ufds[f].revents & POLLIN) {
 						int client;
 						struct sockaddr addr;
@@ -1037,7 +1191,8 @@ dispatch_runner(void *arg)
 						for (c = 0; c < MAX_LISTENERS; c++) {
 							if (listeners[c] == NULL)
 								continue;
-							for (sock = listeners[c]->socks; *sock != -1; sock++) {
+							sock = listeners[c]->socks;
+							for ( ; *sock != -1; sock++) {
 								if (ufds[f].fd == *sock)
 									break;
 							}
@@ -1051,7 +1206,8 @@ dispatch_runner(void *arg)
 							close(client);
 							continue;
 						}
-						if (dispatch_addconnection(client, listeners[c]) == -1) {
+						if (dispatch_addconnection(client, listeners[c]) == -1)
+						{
 							close(client);
 							continue;
 						}
@@ -1059,17 +1215,24 @@ dispatch_runner(void *arg)
 				}
 			}
 		}
+
+		if (ufds != NULL)
+			free(ufds);
 	} else if (self->type == CONNECTION) {
 		int work;
-		struct timeval start, stop;
+		struct timeval start;
+		struct timeval stop;
 
 		while (__sync_bool_compare_and_swap(&(self->keep_running), 1, 1)) {
 			work = 0;
 
-			if (__sync_bool_compare_and_swap(&(self->route_refresh_pending), 1, 1)) {
+			if (__sync_bool_compare_and_swap(&(self->route_refresh_pending),
+						1, 1))
+			{
 				self->rtr = self->pending_rtr;
 				self->pending_rtr = NULL;
-				__sync_bool_compare_and_swap(&(self->route_refresh_pending), 1, 0);
+				__sync_bool_compare_and_swap(&(self->route_refresh_pending),
+						1, 0);
 				__sync_and_and_fetch(&(self->hold), 0);
 			}
 
@@ -1099,7 +1262,11 @@ dispatch_runner(void *arg)
 					work == 0)
 			{
 				gettimeofday(&start, NULL);
-				usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
+				/* wait a bit, but immediately spurt into action if
+				 * there's data available */
+				if (sema_wait(datawaiting,  /* 700ms - 999ms */
+							(700 + (rand() % 1000)) * 1000000) == 0)
+					tracef("dispatcher %d woken up\n", self->id);
 				gettimeofday(&stop, NULL);
 				__sync_add_and_fetch(&(self->sleeps), timediff(start, stop));
 			}
@@ -1120,7 +1287,9 @@ dispatch_new(
 		unsigned char id,
 		enum conntype type,
 		router *r,
-		char *allowed_chars
+		char *allowed_chars,
+		int maxinplen,
+		int maxmetriclen
 	)
 {
 	dispatcher *ret = malloc(sizeof(dispatcher));
@@ -1132,6 +1301,11 @@ dispatch_new(
 		return NULL;
 	}
 
+	if (type == CONNECTION && datawaiting == NULL) {
+		datawaiting = &_datawaiting;
+		sem_init(datawaiting, 0, 0);
+	}
+
 	ret->id = id + 1;  /* ensure > 0 */
 	ret->type = type;
 	ret->keep_running = 1;
@@ -1140,9 +1314,12 @@ dispatch_new(
 	ret->hold = 0;
 	ret->allowed_chars = allowed_chars;
 	ret->tags_supported = 0;
+	ret->maxinplen = maxinplen;
+	ret->maxmetriclen = maxmetriclen;
 
 	ret->metrics = 0;
 	ret->blackholes = 0;
+	ret->discards = 0;
 	ret->ticks = 0;
 	ret->sleeps = 0;
 	ret->prevmetrics = 0;
@@ -1194,7 +1371,7 @@ dispatch_init_listeners()
 dispatcher *
 dispatch_new_listener(unsigned char id)
 {
-	return dispatch_new(id, LISTENER, NULL, NULL);
+	return dispatch_new(id, LISTENER, NULL, NULL, 0, 0);
 }
 
 /**
@@ -1202,9 +1379,11 @@ dispatch_new_listener(unsigned char id)
  * existing connections.
  */
 dispatcher *
-dispatch_new_connection(unsigned char id, router *r, char *allowed_chars)
+dispatch_new_connection(unsigned char id, router *r, char *allowed_chars,
+		int maxinplen, int maxmetriclen)
 {
-	return dispatch_new(id, CONNECTION, r, allowed_chars);
+	return dispatch_new(id, CONNECTION, r, allowed_chars,
+			maxinplen, maxmetriclen);
 }
 
 /**
@@ -1356,6 +1535,27 @@ dispatch_get_blackholes_sub(dispatcher *self)
 	return d;
 }
 
+/**
+ * Returns the number of metrics that were discarded since start.
+ */
+inline size_t
+dispatch_get_discards(dispatcher *self)
+{
+	return __sync_add_and_fetch(&(self->discards), 0);
+}
+
+/**
+ * Returns the number of metrics that were discarded since last call to
+ * this function.
+ */
+inline size_t
+dispatch_get_discards_sub(dispatcher *self)
+{
+	size_t d = dispatch_get_discards(self) - self->prevdiscards;
+	self->prevdiscards += d;
+	return d;
+}
+
 /**
  * Returns the number of accepted connections thusfar.
  */
