@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2018 Fabian Groffen
+ * Copyright 2013-2020 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -308,6 +308,9 @@ handle_signal(int sig)
 		case SIGHUP:
 			sign = "SIGHUP";
 			break;
+		case SIGUSR1:
+			sign = "SIGUSR1";
+			break;
 	}
 	if (keep_running) {
 		logout("caught %s\n", sign);
@@ -319,6 +322,12 @@ handle_signal(int sig)
 	}
 	if (sig == SIGHUP) {
 		do_reload();
+	} else if (sig == SIGUSR1) {
+		if (mode & MODE_DEBUG) {
+			mode &= ~MODE_DEBUG;
+		} else {
+			mode |= MODE_DEBUG;
+		}
 	} else {
 		keep_running = 0;
 	}
@@ -393,13 +402,14 @@ do_usage(char *name, int exitcode)
 #ifdef HAVE_SSL
 	printf("  -C  use CA <cert> to verify SSL connections\n");
 #endif
-	printf("  -S  statistics sending interval in seconds, defaults to 60\n");
 	printf("  -B  connection listen backlog, defaults to 32\n");
 	printf("  -U  socket receive buffer size, max/min/default values depend on OS\n");
 	printf("  -T  IO timeout in milliseconds for server connections, defaults to %d\n", iotimeout);
 	printf("  -E  disable disconnecting idle connections after 10 minutes\n");
-	printf("  -m  send statistics like carbon-cache.py, e.g. not cumulative\n");
 	printf("  -c  characters to allow next to [A-Za-z0-9], defaults to -_:#\n");
+	printf("  -m  max string length of metric, defaults to 0, limit of -M\n");
+	printf("  -M  max string length of metric+ts+value+nl, defaults to %d\n",
+			METRIC_BUFSIZ);
 	printf("  -d  debug mode: currently writes statistics to log, prints hash\n"
 	       "      ring contents and matching position in test mode (-t)\n");
 #ifdef ENABLE_TRACE
@@ -409,7 +419,7 @@ do_usage(char *name, int exitcode)
 	       "      statistics, report drop counts and queue pressure to log\n");
 	printf("  -t  config test mode: prints rule matches from input on stdin\n");
 	printf("  -H  hostname: override hostname (used in statistics)\n");
-	printf("  -D  daemonise: run in a background\n");
+	printf("  -D  daemonise: detach and run in background\n");
 	printf("  -P  pidfile: write a pid to a specified pidfile\n");
 	printf("  -O  minimum number of rules before optimising the ruleset, default: %d\n", optimiserthreshold); 
 
@@ -429,12 +439,14 @@ main(int argc, char * const argv[])
 	int fds[2];
 	unsigned char startup_success = 0;
 	listener *lsnrs;
+	int maxmetriclen = -1;
+	int maxinplen = -1;
 
 	if (gethostname(relay_hostname, sizeof(relay_hostname)) < 0)
 		snprintf(relay_hostname, sizeof(relay_hostname), "127.0.0.1");
 
 	while ((ch = getopt(argc, argv,
-					":hvdstf:l:p:w:b:q:L:C:T:c:H:B:U:EDP:O:")) != -1)
+					":hvdstf:l:p:w:b:q:L:C:T:c:m:M:H:B:U:EDP:O:")) != -1)
 	{
 		switch (ch) {
 			case 'v':
@@ -497,7 +509,9 @@ main(int argc, char * const argv[])
 				break;
 			case 'L':
 				maxstalls = atoi(optarg);
-				if (maxstalls < 0 || maxstalls >= (1 << SERVER_STALL_BITS)) {
+				if (maxstalls < 0 || maxstalls >= (1 << SERVER_STALL_BITS) ||
+						!isdigit(*optarg))
+				{
 					fprintf(stderr, "error: maximum stalls needs to be a number "
 							"between 0 and %d\n", (1 << SERVER_STALL_BITS) - 1);
 					do_usage(argv[0], 1);
@@ -527,6 +541,29 @@ main(int argc, char * const argv[])
 			case 'c':
 				allowed_chars = optarg;
 				break;
+			case 'm': {
+				int val = atoi(optarg);
+				if (val < 0 || !isdigit(*optarg)) {
+					fprintf(stderr, "error: max metric length needs to "
+							"be a number >=0\n");
+					do_usage(argv[0], 1);
+				}
+				maxmetriclen = val;
+			}	break;
+			case 'M': {
+				int val = atoi(optarg);
+				if (val <= 0) {
+					fprintf(stderr, "error: max length needs to "
+							"be a number >0\n");
+					do_usage(argv[0], 1);
+				}
+				if (val > METRIC_BUFSIZ) {
+					fprintf(stderr, "error: max length cannot be >%d\n",
+							METRIC_BUFSIZ);
+					do_usage(argv[0], 1);
+				}
+				maxinplen = val;
+			}	break;
 			case 'H':
 				snprintf(relay_hostname, sizeof(relay_hostname), "%s", optarg);
 				break;
@@ -582,6 +619,11 @@ main(int argc, char * const argv[])
 				break;
 			case 'O': {
 				int val = atoi(optarg);
+				if (val == 0 && !isdigit(*optarg)) {
+					fprintf(stderr, "error: optimiser threshold needs to "
+							"be a number\n");
+					do_usage(argv[0], 1);
+				}
 				optimiserthreshold = val < 0 ? -1 : val;
 			}	break;
 			case '?':
@@ -605,7 +647,7 @@ main(int argc, char * const argv[])
 		workercnt = mode & MODE_SUBMISSION ? 2 : get_cores();
 
 	/* disable collector for submission mode */
-	if (mode & MODE_SUBMISSION)
+	if (mode & MODE_SUBMISSION && !(mode & MODE_DEBUG))
 		collector_interval = 0;
 
 	/* any_of failover maths need batchsize to be smaller than queuesize */
@@ -614,6 +656,18 @@ main(int argc, char * const argv[])
 		exit(1);
 	}
 
+	/* check bufsizes make sense */
+	if (maxinplen > 0 && maxmetriclen > 0 && maxmetriclen >= maxinplen)
+	{
+		fprintf(stderr, "error: max metric len must be smaller than "
+				"max length of input string\n");
+		exit(1);
+	}
+	if (maxinplen == -1)
+		maxinplen = METRIC_BUFSIZ;
+	if (maxmetriclen == -1)
+		maxmetriclen = maxinplen;  /* never hit the limit */
+
 	if (relay_logfile != NULL && (mode & MODE_TEST) == 0) {
 		FILE *f = fopen(relay_logfile, "a");
 		if (f == NULL) {
@@ -794,6 +848,10 @@ main(int argc, char * const argv[])
 		if (allowed_chars != NULL)
 			fprintf(relay_stdout, "    extra allowed characters = %s\n",
 					allowed_chars);
+		if (maxinplen < METRIC_BUFSIZ)
+			fprintf(relay_stdout, "    max input length = %d\n", maxinplen);
+		if (maxmetriclen < maxinplen)
+			fprintf(relay_stdout, "    max metric length = %d\n", maxmetriclen);
 		if (mode & MODE_DEBUG)
 			fprintf(relay_stdout, "    debug = true\n");
 #ifdef ENABLE_TRACE
@@ -875,6 +933,9 @@ main(int argc, char * const argv[])
 	if (signal(SIGHUP, sig_handler) == SIG_ERR) {
 		exit_err("failed to create SIGHUP handler: %s\n", strerror(errno));
 	}
+	if (signal(SIGUSR1, sig_handler) == SIG_ERR) {
+		exit_err("failed to create SIGUSR1 handler: %s\n", strerror(errno));
+	}
 	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
 		exit_err("failed to ignore SIGPIPE: %s\n", strerror(errno));
 	}
@@ -908,7 +969,8 @@ main(int argc, char * const argv[])
 	logout("starting %d workers\n", workercnt);
 	for (id = 0; id < workercnt; id++) {
 		workers[id + 1] = dispatch_new_connection(
-				(unsigned char)id, rtr, allowed_chars);
+				(unsigned char)id, rtr, allowed_chars,
+				maxinplen, maxmetriclen);
 		if (workers[id + 1] == NULL) {
 			logerr("failed to add worker %d\n", id);
 			break;
