@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2018 Fabian Groffen
+ * Copyright 2013-2019 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -395,10 +395,10 @@ router_yyerror(
 char *
 router_validate_address(
 		router *rtr,
-		char **retip, int *retport, void **retsaddr, void **rethint,
-		char *ip, con_proto proto)
+		char **retip, unsigned short *retport, void **retsaddr,
+		void **rethint, char *ip, con_proto proto)
 {
-	int port = GRAPHITE_PORT;
+	unsigned short port = GRAPHITE_PORT;
 	struct addrinfo *saddr = NULL;
 	struct addrinfo hint;
 	char sport[8];
@@ -412,11 +412,17 @@ router_validate_address(
 	if (*(p - 1) == ']')
 		lastcolon = NULL;
 	if (lastcolon != NULL) {
+		long lport;
 		char *endp = NULL;
 		*lastcolon = '\0';
-		port = (int)strtol(lastcolon + 1, &endp, 10);
-		if (port < 1 || endp != p)
-			return(ra_strdup(rtr->a, "invalid port number"));
+		lport = strtol(lastcolon + 1, &endp, 10);
+		if (lport < 1 || lport > UINT16_MAX || endp != p) {
+			char errbuf[40];
+			snprintf(errbuf, sizeof(errbuf),
+				 "invalid port number '%s'", lastcolon + 1);
+			return(ra_strdup(rtr->a, errbuf));
+		}
+		port = lport;
 	}
 	if (*ip == '[') {
 		ip++;
@@ -579,7 +585,7 @@ char *
 router_add_server(
 		router *ret,
 		char *ip,
-		int port,
+		unsigned short port,
 		char *inst,
 		con_type type,
 		con_trnsp transport,
@@ -625,7 +631,7 @@ router_add_server(
 			}
 		}
 		if (newserver == NULL) {
-			newserver = server_new(ip, (unsigned short)port,
+			newserver = server_new(ip, port,
 					type, transport, proto, walk, hint,
 					ret->conf.queuesize, ret->conf.batchsize,
 					ret->conf.maxstalls, ret->conf.iotimeout,
@@ -859,8 +865,11 @@ router_add_cluster(router *r, cluster *cl)
 		 * number of servers */
 		size_t i = 0;
 		char errbuf[512];
+		server **secondaries;
+
 		for (w = cl->members.ch->servers; w != NULL; w = w->next)
 			i++;
+
 		if (i < cl->members.ch->repl_factor) {
 			snprintf(errbuf, sizeof(errbuf),
 					"invalid cluster '%s': replication count (%u) is "
@@ -868,6 +877,16 @@ router_add_cluster(router *r, cluster *cl)
 					cl->name, cl->members.ch->repl_factor, i);
 			return ra_strdup(r->a, errbuf);
 		}
+
+		if (cl->isdynamic) {
+			secondaries = ra_malloc(r->a, sizeof(server *) * i);
+			if (secondaries == NULL)
+				return ra_strdup(r->a, "malloc failed for dynamic servers");
+			for (i = 0, w = cl->members.ch->servers; w != NULL; w = w->next)
+				secondaries[i++] = w->server;
+			for (w = cl->members.ch->servers; w != NULL; w = w->next)
+				server_add_secondaries(w->server, secondaries, i);
+		}
 	}
 	return NULL;
 }
@@ -996,7 +1015,7 @@ router_add_listener(
 		char *pemcert,
 		con_proto ctype,
 		char *ip,
-		int port,
+		unsigned short port,
 		struct addrinfo *saddrs)
 {
 	listener *lwalk;
@@ -1296,7 +1315,7 @@ router_readconfig(router *orig,
 			char *p;
 			char *carets;
 			size_t carlen;
-			fprintf(stderr, "%s:%zd:%zd: %s\n", path, ret->parser_err.line,
+			logerr("%s:%zd:%zd: %s\n", path, ret->parser_err.line,
 					ret->parser_err.start, ret->parser_err.msg);
 			/* get some relevant context from buff and put ^^^^ below it
 			 * to point out the position of the error */
@@ -1314,15 +1333,15 @@ router_readconfig(router *orig,
 			carets = ra_malloc(palloc, carlen + 1);
 			memset(carets, '^', carlen);
 			carets[carlen] = '\0';
-			fprintf(stderr, "%s\n", line);
+			logerr("%s\n", line);
 			/* deal with tabs in the input */
 			for (p = line; p - line < ret->parser_err.start; p++)
 				if (*p != '\t')
 					*p = ' ';
 			*p = '\0';
-			fprintf(stderr, "%s%s\n", line, carets);
+			logerr("%s%s\n", line, carets);
 		} else if (ret->parser_err.msg != NULL) {
-			fprintf(stderr, "%s: %s\n", path, ret->parser_err.msg);
+			logerr("%s: %s\n", path, ret->parser_err.msg);
 		}
 		/* clear the message, so it isn't printed when this was an
 		 * included config */
@@ -1340,7 +1359,7 @@ router_readconfig(router *orig,
 		if (ret->listeners == NULL) {
 			char sockbuf[128];
 			char *ip;
-			int port;
+			unsigned short port;
 			void *saddrs;
 			void *hint;
 
@@ -1586,9 +1605,10 @@ router_optimise(router *r, int threshold)
 			rnext = rwalk->next;
 			rwalk->next = NULL;
 			if (rwalk->stop) {
-				bstart = blast;
+				bstart = NULL;
 				rlast = NULL;
 				seq++;
+				tracef("terminating group due to stop rule\n");
 			}
 			continue;
 		}
@@ -1607,6 +1627,9 @@ router_optimise(router *r, int threshold)
 					bwalk->next = NULL;
 				}
 				rlast = rwalk;
+				bstart = NULL;
+				seq++;  /* finish this group */
+				tracef("finishing group due to stop rule\n");
 			}
 		} else {
 			bstart = blast;
@@ -1621,14 +1644,21 @@ router_optimise(router *r, int threshold)
 	blast->next = NULL;
 	blast->seqnr = seq;
 
+	tracef("reducing into groups\n");
 	rwalk = r->routes = NULL;
 	seq = 1;
 	/* create groups, if feasible */
 	for (bwalk = blocks; bwalk != NULL; bwalk = blast) {
-		if (bwalk->seqnr != seq) {
+		tracef("seq=%zd, bwalk=%p, ->seqnr=%zd, ->refcnt=%zd, ->pattern=%s\n",
+				seq, bwalk, bwalk->seqnr, bwalk->refcnt, bwalk->pattern);
+		if (bwalk->seqnr > seq) {
 			seq++;
 			blast = bwalk;
 			continue;
+		} else if (bwalk->seqnr < seq) {
+			logerr("routes optimisation lost track, "
+					"skipping further optimisations\n");
+			return;
 		}
 
 		if (bwalk->refcnt == 0) {
@@ -1787,6 +1817,58 @@ router_set_collectorvals(router *rtr, int intv, char *prefix, col_mode smode)
 	return NULL;
 }
 
+static char *_router_quoteident_buf = NULL;
+static size_t _router_quoteident_buflen = 0;
+static const char *router_quoteident(const char *ident)
+{
+	/*
+	IDCHR   [a-zA-Z0-9[\](){}|\\^?*$&%<>:/,._+-]
+	ident = {IDCHR}+(\\[ ]{IDCHR}+)*
+	*/
+	char allowodd[] = { '[', ']', '(', ')', '{', '}',
+		'|', '\\', '^', '?', '*', '$', '%', '<', '>',
+		':', '/', ',', '.', '_', '+', '-', '\0' /* string terminator */};
+	size_t idlen = strlen(ident);
+	size_t idpos;
+	size_t bufpos;
+	char needquote = 0;
+
+	/* parser won't accept empty strings for input, so we can assume
+	 * idlen >= 1 here */
+
+	/* setup buf for worst case, allocate in 32-byte blocks to try and
+	 * avoid reallocs for just a few bytes */
+	if (_router_quoteident_buflen < 1 + (idlen * 2) + 1 + 1) {
+		_router_quoteident_buflen = (((1 + (idlen * 2) + 1 + 1) / 32) + 1) * 32;
+		_router_quoteident_buf =
+			realloc(_router_quoteident_buf, _router_quoteident_buflen);
+	}
+
+	bufpos = 0;
+	_router_quoteident_buf[bufpos++] = '"';
+
+	for (idpos = 0; idpos < idlen; idpos++) {
+		if (!(
+				(ident[idpos] >= 'a' && ident[idpos] <= 'z') ||
+				(ident[idpos] >= 'A' && ident[idpos] <= 'Z') ||
+				(ident[idpos] >= '0' && ident[idpos] <= '9') ||
+				strchr(allowodd, (int)ident[idpos]) != NULL)
+		   )
+		{
+			if (!(idpos > 1 && ident[idpos] == ' ' && ident[idpos - 1] == '\\'))
+				needquote = 1;
+		}
+		_router_quoteident_buf[bufpos++] = ident[idpos];
+	}
+	_router_quoteident_buf[bufpos++] = '"';
+	_router_quoteident_buf[bufpos] = '\0';
+
+	if (needquote)
+		return _router_quoteident_buf;
+
+	return ident;
+}
+
 /**
  * Mere debugging function to check if the configuration is picked up
  * alright.  If all is set to false, aggregation rules won't be printed.
@@ -1826,7 +1908,8 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 							con_trnsp_str[walk->transport & 0xFFFF],
 						(walk->transport & ~0xFFFF) == W_SSL ? " ssl " : "",
 #ifdef HAVE_SSL
-						(walk->transport & ~0xFFFF) == W_SSL ? walk->pemcert : ""
+						(walk->transport & ~0xFFFF) == W_SSL ?
+						router_quoteident(walk->pemcert) : ""
 #else
 						""
 #endif
@@ -1842,7 +1925,8 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 				fprintf(f, "\n");
 				do {
 					if (walk->ctype == CON_UNIX) {
-						fprintf(f, "        %s proto unix\n", walk->ip);
+						fprintf(f, "        %s proto unix\n",
+								router_quoteident(walk->ip));
 					} else {
 						fprintf(f, "        %s%s%u proto %s\n",
 								walk->ip == NULL ? "" : walk->ip,
@@ -1862,17 +1946,19 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 				rtr->collector.interval);
 		if (rtr->collector.mode == SUB)
 			fprintf(f, "    reset counters after interval\n");
-		fprintf(f, "    prefix with %s\n", rtr->collector.prefix);
+		fprintf(f, "    prefix with %s\n",
+				router_quoteident(rtr->collector.prefix));
 		if (rtr->collector.stub != NULL) {
 			fprintf(f, "    send to");
 			for (r = rtr->routes; r != NULL; r = r->next) {
 				if (r->dests->cl->type == STATSTUB) {
 					destinations *d = r->dests->cl->members.routes->dests;
 					if (d->next == NULL) {
-						fprintf(f, " %s", d->cl->name);
+						fprintf(f, " %s", router_quoteident(d->cl->name));
 					} else {
 						for ( ; d != NULL; d = d->next)
-							fprintf(f, "\n        %s", d->cl->name);
+							fprintf(f, "\n        %s",
+									router_quoteident(d->cl->name));
 					}
 					fprintf(f, "\n    stop\n");
 					break;
@@ -1889,7 +1975,7 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 				c->type == GROUP || c->type == AGGREGATION ||
 				c->type == AGGRSTUB || c->type == STATSTUB)
 			continue;
-		fprintf(f, "cluster %s\n", c->name);
+		fprintf(f, "cluster %s\n", router_quoteident(c->name));
 		if (c->type == FORWARD) {
 			fprintf(f, "    forward\n");
 			for (s = c->members.forward; s != NULL; s = s->next)
@@ -1900,7 +1986,7 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			fprintf(f, "    file%s\n", c->type == FILELOGIP ? " ip" : "");
 			for (s = c->members.forward; s != NULL; s = s->next)
 				fprintf(f, "        %s\n",
-						serverip(s->server));
+						router_quoteident(serverip(s->server)));
 		} else if (c->type == ANYOF || c->type == FAILOVER) {
 			fprintf(f, "    %s\n", c->type == ANYOF ? "any_of" : "failover");
 			for (s = c->members.anyof->list; s != NULL; s = s->next)
@@ -1911,10 +1997,11 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 				c->type == FNV1A_CH ||
 				c->type == JUMP_CH)
 		{
-			fprintf(f, "    %s_ch replication %d\n",
+			fprintf(f, "    %s_ch replication %d%s\n",
 					c->type == CARBON_CH ? "carbon" : 
 					c->type == FNV1A_CH ? "fnv1a" : "jump_fnv1a",
-					c->members.ch->repl_factor);
+					c->members.ch->repl_factor,
+					c->isdynamic ? " dynamic" : "");
 			for (s = c->members.ch->servers; s != NULL; s = s->next)
 				fprintf(f, "        %s:%d%s%s%s" PTYPEFMT PTRNSPFMT "\n",
 						serverip(s->server), server_port(s->server),
@@ -1954,11 +2041,11 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 
 			fprintf(f, "aggregate");
 			if (r->next == NULL || r->next->dests->cl != aggr) {
-				fprintf(f, " %s\n", r->pattern);
+				fprintf(f, " %s\n", router_quoteident(r->pattern));
 			} else {
 				fprintf(f, "\n");
 				do {
-					fprintf(f, "        %s\n", r->pattern);
+					fprintf(f, "        %s\n", router_quoteident(r->pattern));
 				} while (r->next != NULL && r->next->dests->cl == aggr
 						&& (r = r->next) != NULL);
 			}
@@ -1971,7 +2058,8 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 					aggr->members.aggregation->tswhen == TS_MIDDLE ? "middle" :
 					aggr->members.aggregation->tswhen == TS_END ? "end" :
 					"<unknown>");
-			for (ac = aggr->members.aggregation->computes; ac != NULL; ac = ac->next) {
+			ac = aggr->members.aggregation->computes;
+			for ( ; ac != NULL; ac = ac->next) {
 				snprintf(percentile, sizeof(percentile),
 						"percentile%d", ac->percentile);
 				fprintf(f, "    compute %s write to\n"
@@ -1984,23 +2072,25 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 						ac->type == VAR ? "variance" :
 						ac->type == SDEV ? "stddev" :
 						"<unknown>",
-						ac->metric + strlen(stubname));
+						router_quoteident(ac->metric + strlen(stubname)));
 			}
 			if (!(pmode & PMODE_STUB) && r->dests->next != NULL) {
 				destinations *dn = r->dests->next;
 				fprintf(f, "    send to");
 				if (dn->next == NULL) {
-					fprintf(f, " %s\n", dn->cl->name);
+					fprintf(f, " %s\n", router_quoteident(dn->cl->name));
 				} else {
 					for (; dn != NULL; dn = dn->next)
-						fprintf(f, "\n        %s", dn->cl->name);
+						fprintf(f, "\n        %s",
+								router_quoteident(dn->cl->name));
 					fprintf(f, "\n");
 				}
 			}
 			fprintf(f, "%s    ;\n", r->stop ? "    stop\n" : "");
 		} else if (r->dests->cl->type == REWRITE) {
-			fprintf(f, "rewrite %s\n    into %s\n    ;\n",
-					r->pattern, r->dests->cl->members.replacement);
+			fprintf(f, "rewrite %s\n", router_quoteident(r->pattern));
+			fprintf(f, "    into %s\n    ;\n",
+					router_quoteident(r->dests->cl->members.replacement));
 		} else if (r->dests->cl->type == GROUP) {
 			size_t cnt = 0;
 			route *rwalk;
@@ -2008,7 +2098,8 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			char *b = &blockname[sizeof(blockname) - 1];
 			char *p;
 
-			for (rwalk = r->dests->cl->members.routes; rwalk != NULL; rwalk = rwalk->next)
+			rwalk = r->dests->cl->members.routes;
+			for ( ; rwalk != NULL; rwalk = rwalk->next)
 				cnt++;
 			/* reverse the name, to make it human consumable */
 			*b-- ='\0';
@@ -2031,14 +2122,16 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			if (pmode & PMODE_STUB) {
 				fprintf(f, "# stub match for aggregate/statistics rule "
 						"with send to\n");
-				fprintf(f, "match ^%s\n    send to", r->pattern);
+				fprintf(f, "match ^%s\n    send to",
+						router_quoteident(r->pattern));
 				if (r->dests->cl->members.routes->dests->next == NULL) {
-					fprintf(f, " %s",
-							r->dests->cl->members.routes->dests->cl->name);
+					fprintf(f, " %s", router_quoteident(
+							r->dests->cl->members.routes->dests->cl->name));
 				} else {
 					destinations *d = r->dests->cl->members.routes->dests;
 					for (; d != NULL; d = d->next)
-						fprintf(f, "\n        %s", d->cl->name);
+						fprintf(f, "\n        %s",
+								router_quoteident(d->cl->name));
 				}
 				fprintf(f, "%s\n    ;\n", r->stop ? "\n    stop" : "");
 			}
@@ -2048,12 +2141,14 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			fprintf(f, "match");
 			if (r->next == NULL || r->next->dests != or->dests) {
 				fprintf(f, " %s\n",
-						r->matchtype == MATCHALL ? "*" : r->pattern);
+						r->matchtype == MATCHALL ? "*" :
+						router_quoteident(r->pattern));
 			} else {
 				fprintf(f, "\n");
 				do {
 					fprintf(f, "        %s\n",
-							r->matchtype == MATCHALL ? "*" : r->pattern);
+							r->matchtype == MATCHALL ? "*" :
+							router_quoteident(r->pattern));
 				} while (r->next != NULL && r->next->dests == or->dests
 						&& (r = r->next) != NULL);
 			}
@@ -2061,20 +2156,21 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			if (d->cl->type == VALIDATION) {
 				validate *v = d->cl->members.validation;
 				fprintf(f, "    validate %s else %s\n",
-						v->rule->pattern,
+						router_quoteident(v->rule->pattern),
 						v->action == VAL_LOG ? "log" : "drop");
 				/* hide this pseudo target */
 				d = d->next;
 			}
 			if (r->masq != NULL)
-				fprintf(f, "    route using %s\n", r->masq);
+				fprintf(f, "    route using %s\n", router_quoteident(r->masq));
 			if (d != NULL) {
 				fprintf(f, "    send to");
 				if (d->next == NULL) {
-					fprintf(f, " %s", d->cl->name);
+					fprintf(f, " %s", router_quoteident(d->cl->name));
 				} else {
 					for ( ; d != NULL; d = d->next)
-						fprintf(f, "\n        %s", d->cl->name);
+						fprintf(f, "\n        %s",
+								router_quoteident(d->cl->name));
 				}
 				fprintf(f, "\n%s    ;\n", or->stop ? "    stop\n" : "");
 			} else {
@@ -2098,7 +2194,7 @@ router_printdiffs(router *old, router *new, FILE *out)
 	char *tmp = getenv("TMPDIR");
 	char patho[512];
 	char pathn[512];
-	char buf[1024];
+	char buf[1033];
 	size_t len;
 	int ret;
 	mode_t mask;
@@ -2147,6 +2243,10 @@ router_printdiffs(router *old, router *new, FILE *out)
 	/* diff and print its output */
 	snprintf(buf, sizeof(buf), "diff -u %s %s", patho, pathn);
 	f = popen(buf, "r");
+	if (f == NULL) {
+		logerr("failed to popen diff: %s\n", strerror(errno));
+		return 1;
+	}
 	while ((len = fread(buf, 1, sizeof(buf) - 1, f)) > 0) {
 		if (fwrite(buf, len, 1, out) == 0)
 			/* ignore */
@@ -2859,6 +2959,10 @@ router_route_intern(
 								firstspace,
 								w->dests->cl->members.routes,
 								dispatcher_id);
+						/* ensure recursion doesn't result in false
+						 * blachole stats #405 */
+						if (*blackholed == 0)
+							wassent = 1;
 					}	break;
 					case VALIDATION: {
 						/* test whether data matches, if not, either log
@@ -3022,8 +3126,8 @@ router_test_intern(char *metric, char *firstspace, route *routes)
 						if (mode & MODE_DEBUG || d->next == NULL) {
 							stublen = 0;
 						} else {
-							char x;
-							stublen = snprintf(&x, 1,
+							char x[32];
+							stublen = snprintf(x, 32,
 									STUB_AGGR "%p__",
 									d->cl->members.aggregation);
 						}
