@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2018 Fabian Groffen
+ * Copyright 2013-2019 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -44,6 +44,7 @@
 #endif
 #ifdef HAVE_LZ4
 #include <lz4.h>
+#include <lz4frame.h>
 #endif
 #ifdef HAVE_SNAPPY
 #include <snappy-c.h>
@@ -71,7 +72,10 @@ typedef struct _z_strm {
 		z_streamp gz;
 #endif
 #ifdef HAVE_LZ4
-		LZ4_stream_t *lz;
+		struct {
+			void *cbuf;
+			size_t cbuflen;
+		} z;
 #endif
 #ifdef HAVE_SSL
 		SSL *ssl;
@@ -243,24 +247,37 @@ gziperror(z_strm *strm, int rval)
 
 #ifdef HAVE_LZ4
 /* lz4 wrapped socket */
+
+static inline int lzflush(z_strm *strm);
+
 static inline ssize_t
 lzwrite(z_strm *strm, const void *buf, size_t sze)
 {
-	int oret;
-	char *obuf = strm->obuf;
-	int cret;
+	size_t towrite = sze;
 
-	cret = LZ4_compress_fast_continue(strm->hdl.lz,
-			buf, strm->obuf, sze, METRIC_BUFSIZ, 0);
-	if (cret == 0)
-		return -1;  /* we must reset/free lz */
-	while (cret > 0) {
-		oret = strm->nextstrm->strmwrite(strm->nextstrm, strm->obuf, cret);
-		if (oret < 0)
-			return -1;  /* failure is failure */
-		/* update counters to possibly retry the remaining bit */
-		obuf += oret;
-		cret -= oret;
+	/* use the same strategy as gzip: fill the buffer until space
+	   runs out. we completely fill the output buffer before flushing */
+
+	while (towrite > 0) {
+
+		size_t avail = METRIC_BUFSIZ - strm->obuflen;
+		size_t copysize = towrite > avail ? avail : towrite;
+
+		/* copy into the output buffer as much as we can */
+
+		if (copysize > 0) {
+			memcpy(strm->obuf + strm->obuflen, buf, copysize);
+			strm->obuflen += copysize;
+			towrite -= copysize;
+			buf += copysize;
+		}
+
+		/* if output buffer is full & still have bytes to write, flush now */
+
+		if (strm->obuflen == METRIC_BUFSIZ && towrite > 0 && lzflush(strm) != 0) {
+			logerr("Failed to flush LZ4 data to make space\n");
+			return -1;
+		}
 	}
 
 	return sze;
@@ -269,15 +286,45 @@ lzwrite(z_strm *strm, const void *buf, size_t sze)
 static inline int
 lzflush(z_strm *strm)
 {
-	return strm->nextstrm->strmflush(strm->nextstrm);
+	int oret;
+	size_t ret;
+
+	/* anything to do? */
+
+	if (strm->obuflen == 0)
+		return 0;
+
+	/* the buffered data goes out as a single frame */
+
+	ret = LZ4F_compressFrame(strm->hdl.z.cbuf, strm->hdl.z.cbuflen, strm->obuf, strm->obuflen, NULL);
+	if (LZ4F_isError(ret)) {
+		logerr("Failed to compress %d LZ4 bytes into a frame: %d\n",
+		       strm->obuflen, (int)ret);
+		return -1;
+	}
+		/* write and flush */
+
+	if ((oret = strm->nextstrm->strmwrite(strm->nextstrm,
+					      strm->hdl.z.cbuf, ret)) < 0) {
+		logerr("Failed to write %lu bytes of compressed LZ4 data: %d\n",
+		       ret, oret);
+		return oret;
+	}
+
+	strm->nextstrm->strmflush(strm->nextstrm);
+
+	/* the buffer is gone. reset the write position */
+
+	strm->obuflen = 0;
+	return 0;
 }
 
 static inline int
 lzclose(z_strm *strm)
 {
-	int ret = strm->nextstrm->strmclose(strm->nextstrm);
-	LZ4_freeStream(strm->hdl.lz);
-	return ret;
+	lzflush(strm);
+	free(strm->hdl.z.cbuf);
+	return strm->nextstrm->strmclose(strm->nextstrm);
 }
 
 static inline const char *
@@ -373,11 +420,53 @@ sslclose(z_strm *strm)
 	return close(sock);
 }
 
+static char _sslerror_buf[256];
 static inline const char *
 sslerror(z_strm *strm, int rval)
 {
 	int err = SSL_get_error(strm->hdl.ssl, rval);
-	return ERR_reason_error_string(err);
+	switch (err) {
+		case SSL_ERROR_NONE:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: SSL_ERROR_NONE", err);
+			break;
+		case SSL_ERROR_ZERO_RETURN:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: TLS/SSL connection has been closed", err);
+			break;
+		case SSL_ERROR_WANT_READ:
+		case SSL_ERROR_WANT_WRITE:
+		case SSL_ERROR_WANT_CONNECT:
+		case SSL_ERROR_WANT_ACCEPT:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: the read or write operation did not complete", err);
+			break;
+		case SSL_ERROR_WANT_X509_LOOKUP:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: call callback via SSL_CTX_set_client_cert_cb()", err);
+			break;
+#ifdef SSL_ERROR_WANT_ASYNC
+		case SSL_ERROR_WANT_ASYNC:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: asynchronous engine is still processing data", err);
+			break;
+#endif
+#ifdef SSL_ERROR_WANT_ASYNC_JOB
+		case SSL_ERROR_WANT_ASYNC_JOB:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: no async jobs available in the pool", err);
+			break;
+#endif
+		case SSL_ERROR_SYSCALL:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: I/O error: %s", err, strerror(errno));
+			break;
+		case SSL_ERROR_SSL:
+			ERR_error_string_n(ERR_get_error(),
+					_sslerror_buf, sizeof(_sslerror_buf));
+			break;
+	}
+	return _sslerror_buf;
 }
 #endif
 
@@ -519,7 +608,8 @@ server_queuereader(void *d)
 				__sync_add_and_fetch(&(self->dropped), 1);
 			}
 			gettimeofday(&stop, NULL);
-			self->ticks += timediff(start, stop);
+			__sync_add_and_fetch(&(self->ticks),
+					timediff(start, stop));
 			if (squeue == NULL) {
 				/* we couldn't do anything, take it easy for a bit */
 				if (__sync_add_and_fetch(&(self->failure), 0) > 1) {
@@ -795,9 +885,13 @@ server_queuereader(void *d)
 #endif
 #ifdef HAVE_LZ4
 			if ((self->transport & 0xFFFF) == W_LZ4) {
-				self->strm->hdl.lz = LZ4_createStream();
-				if (self->strm->hdl.lz == NULL) {
-					logerr("failed to create lz4 stream: out of memory\n");
+				self->strm->obuflen = 0;
+
+				/* get the maximum size that should ever be required and allocate for it */
+
+				self->strm->hdl.z.cbuflen = LZ4F_compressFrameBound(sizeof(self->strm->obuf), NULL);
+				if ((self->strm->hdl.z.cbuf = malloc(self->strm->hdl.z.cbuflen)) == NULL) {
+					logerr("Failed to allocate %lu bytes for compressed LZ4 data\n", self->strm->hdl.z.cbuflen);
 					close(self->fd);
 					self->fd = -1;
 					continue;
@@ -815,7 +909,7 @@ server_queuereader(void *d)
 				int rv;
 				z_strm *sstrm;
 
-				if (self->transport == W_SSL) { /* just SSL, nothing else */
+				if ((self->transport & 0xFFFF) == W_PLAIN) { /* just SSL, nothing else */
 					sstrm = self->strm;
 				} else {
 					sstrm = self->strm->nextstrm;
@@ -832,16 +926,14 @@ server_queuereader(void *d)
 				}
 				if ((rv = SSL_connect(sstrm->hdl.ssl)) != 1) {
 					logerr("failed to connect ssl stream: %s\n",
-							ERR_reason_error_string(
-								SSL_get_error(sstrm->hdl.ssl, rv)));
+							sslerror(sstrm, rv));
 					sstrm->strmclose(sstrm);
 					self->fd = -1;
 					continue;
 				}
 				if ((rv = SSL_get_verify_result(sstrm->hdl.ssl)) != X509_V_OK) {
 					logerr("failed to verify ssl certificate: %s\n",
-							ERR_reason_error_string(
-								SSL_get_error(sstrm->hdl.ssl, rv)));
+							sslerror(sstrm, rv));
 					sstrm->strmclose(sstrm);
 					self->fd = -1;
 					continue;
@@ -1017,19 +1109,21 @@ server_new(
 		const SSL_METHOD *m = SSLv23_client_method();
 		ret->strm->ctx = SSL_CTX_new(m);
 		
-		if (ret->strm->ctx == NULL ||
-			SSL_CTX_load_verify_locations(ret->strm->ctx,
-										   sslCAisdir ? NULL : sslCA,
-										   sslCAisdir ? sslCA : NULL) == 0)
-		{
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			logerr("failed to create SSL context for server "
-					"%s:%d: %s\n", ret->ip, ret->port, err);
-			free((char *)ret->ip);
-			free(ret->batch);
-			free(ret->strm);
-			free(ret);
-			return NULL;
+		if (sslCA != NULL) {
+			if (ret->strm->ctx == NULL ||
+				SSL_CTX_load_verify_locations(ret->strm->ctx,
+											   sslCAisdir ? NULL : sslCA,
+											   sslCAisdir ? sslCA : NULL) == 0)
+			{
+				char *err = ERR_error_string(ERR_get_error(), NULL);
+				logerr("failed to load SSL verify locations from %s for "
+						"%s:%d: %s\n", sslCA, ret->ip, ret->port, err);
+				free((char *)ret->ip);
+				free(ret->batch);
+				free(ret->strm);
+				free(ret);
+				return NULL;
+			}
 		}
 		SSL_CTX_set_verify(ret->strm->ctx, SSL_VERIFY_PEER, NULL);
 
@@ -1114,8 +1208,6 @@ server_new(
 	}
 	ret->queue = queue_new(qsize);
 	if (ret->queue == NULL) {
-		if (ret->hint)
-			free(ret->hint);
 		free(ret->batch);
 		free((char *)ret->ip);
 		free(ret);
